import struct
from dataclasses import dataclass

from xdsl.backend.riscv.lowering.utils import (
    cast_matched_op_results,
    cast_operands_to_regs,
)
from xdsl.context import Context
from xdsl.dialects import arith, riscv
from xdsl.dialects.builtin import (
    Float32Type,
    Float64Type,
    FloatAttr,
    IndexType,
    IntegerAttr,
    IntegerType,
    ModuleOp,
    UnrealizedConversionCastOp,
)
from xdsl.ir import Operation
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.bitwise_casts import convert_f32_to_u32
from xdsl.utils.comparisons import signed_lower_bound, signed_upper_bound

_INT_REGISTER_TYPE = riscv.Registers.UNALLOCATED_INT
_FLOAT_REGISTER_TYPE = riscv.Registers.UNALLOCATED_FLOAT


class LowerArithConstant(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: arith.ConstantOp, rewriter: PatternRewriter
    ) -> None:
        op_result_type = op.result.type
        if isinstance(op_result_type, IntegerType) and isinstance(
            op_val := op.value, IntegerAttr
        ):
            if op_result_type.width.data <= 32:
                rewriter.replace_matched_op(
                    [
                        constant := riscv.LiOp(op_val.value.data),
                        UnrealizedConversionCastOp.get(
                            constant.results, (op_result_type,)
                        ),
                    ],
                )
            else:
                raise NotImplementedError("Only 32 bit integers are supported for now")
        elif isinstance(op_val := op.value, FloatAttr):
            if isinstance(op_result_type, Float32Type):
                rewriter.replace_matched_op(
                    [
                        lui := riscv.LiOp(
                            convert_f32_to_u32(op_val.value.data),
                            rd=_INT_REGISTER_TYPE,
                        ),
                        fld := riscv.FMvWXOp(lui.rd),
                        UnrealizedConversionCastOp.get(fld.results, (op_result_type,)),
                    ],
                )
            elif isinstance(op_result_type, Float64Type):
                # There is no way to load an immediate value to a float register directly.

                s32_min = signed_lower_bound(32)
                s32_max = signed_upper_bound(32)
                # If the value is an integer that fits in s32, then convert.
                if (val_data := op_val.value.data).is_integer() and s32_min <= (
                    int_val := int(val_data)
                ) < s32_max:
                    rewriter.replace_matched_op(
                        [
                            lui := riscv.LiOp(
                                int_val,
                                rd=_INT_REGISTER_TYPE,
                            ),
                            fcvtdw := riscv.FCvtDWOp(lui.rd, rd=_FLOAT_REGISTER_TYPE),
                            UnrealizedConversionCastOp.get(
                                fcvtdw.results, (op_result_type,)
                            ),
                        ],
                    )
                else:
                    # We have to load the bits into an integer register, store them on the
                    # stack, and load again.

                    # TODO: check the xlen in this lowering.

                    # This lowering assumes that xlen is 32 and flen is 64

                    lower, upper = struct.unpack(
                        "<ii", struct.pack("<d", op_val.value.data)
                    )
                    rewriter.replace_matched_op(
                        [
                            sp := riscv.GetRegisterOp(riscv.Registers.SP),
                            li_upper := riscv.LiOp(upper),
                            riscv.SwOp(sp, li_upper, -4),
                            li_lower := riscv.LiOp(lower),
                            riscv.SwOp(sp, li_lower, -8),
                            fld := riscv.FLdOp(sp, -8, rd=_FLOAT_REGISTER_TYPE),
                            UnrealizedConversionCastOp.get(
                                fld.results, (op_result_type,)
                            ),
                        ],
                    )
            else:
                raise NotImplementedError("Only 32 or 64 bit floats are supported")
        elif isinstance(op_result_type, IndexType) and isinstance(
            op_val := op.value, IntegerAttr
        ):
            rewriter.replace_matched_op(
                [
                    constant := riscv.LiOp(op_val.value.data),
                    UnrealizedConversionCastOp.get(constant.results, (op_result_type,)),
                ],
            )
        else:
            raise NotImplementedError(
                f"Unsupported constant type {op_val} of type {type(op_val)}"
            )


class LowerArithIndexCast(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: arith.IndexCastOp, rewriter: PatternRewriter
    ) -> None:
        """
        On a RV32 triple, the index type is 32 bits, so we can just drop the cast.
        """

        rewriter.replace_matched_op(
            UnrealizedConversionCastOp.get((op.input,), (op.result.type,))
        )


RdRsRsIntegerOperation = riscv.RdRsRsOperation[
    riscv.IntRegisterType, riscv.IntRegisterType, riscv.IntRegisterType
]

RdRsRsFloatOperation = riscv.RdRsRsOperation[
    riscv.FloatRegisterType, riscv.FloatRegisterType, riscv.FloatRegisterType
]


@dataclass
class LowerBinaryIntegerOp(RewritePattern):
    arith_op_cls: type[arith.SignlessIntegerBinaryOperation]
    riscv_op_cls: type[RdRsRsIntegerOperation]

    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter) -> None:
        if not isinstance(op, self.arith_op_cls):
            return

        lhs = UnrealizedConversionCastOp.get((op.lhs,), (_INT_REGISTER_TYPE,))
        rhs = UnrealizedConversionCastOp.get((op.rhs,), (_INT_REGISTER_TYPE,))
        add = self.riscv_op_cls(lhs, rhs, rd=_INT_REGISTER_TYPE)
        cast = UnrealizedConversionCastOp.get((add.rd,), (op.result.type,))

        rewriter.replace_matched_op((lhs, rhs, add, cast))


@dataclass
class LowerBinaryFloatOp(RewritePattern):
    arith_op_cls: type[arith.FloatingPointLikeBinaryOperation]
    riscv_f_op_cls: type[riscv.RdRsRsFloatOperationWithFastMath]
    riscv_d_op_cls: type[riscv.RdRsRsFloatOperationWithFastMath]

    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter) -> None:
        if not isinstance(op, self.arith_op_cls):
            return

        lhs = UnrealizedConversionCastOp.get((op.lhs,), (_FLOAT_REGISTER_TYPE,))
        rhs = UnrealizedConversionCastOp.get((op.rhs,), (_FLOAT_REGISTER_TYPE,))
        match op.lhs.type:
            case Float32Type():
                cls = self.riscv_f_op_cls
            case Float64Type():
                cls = self.riscv_d_op_cls
            case _:
                raise ValueError(f"Unexpected float type {op.lhs.type}")

        rv_flags = riscv.FastMathFlagsAttr(op.fastmath.data)

        new_op = cls(lhs, rhs, rd=_FLOAT_REGISTER_TYPE, fastmath=rv_flags)
        cast = UnrealizedConversionCastOp.get((new_op.rd,), (op.result.type,))

        rewriter.replace_matched_op((lhs, rhs, new_op, cast))


lower_arith_addi = LowerBinaryIntegerOp(arith.AddiOp, riscv.AddOp)
lower_arith_subi = LowerBinaryIntegerOp(arith.SubiOp, riscv.SubOp)
lower_arith_muli = LowerBinaryIntegerOp(arith.MuliOp, riscv.MulOp)
lower_arith_divui = LowerBinaryIntegerOp(arith.DivUIOp, riscv.DivuOp)
lower_arith_divsi = LowerBinaryIntegerOp(arith.DivSIOp, riscv.DivOp)


class LowerArithFloorDivSI(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: arith.FloorDivSIOp, rewriter: PatternRewriter
    ) -> None:
        raise NotImplementedError("FloorDivSI is not supported")


class LowerArithCeilDivSI(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: arith.CeilDivSIOp, rewriter: PatternRewriter
    ) -> None:
        raise NotImplementedError("CeilDivSI is not supported")


class LowerArithCeilDivUI(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: arith.CeilDivUIOp, rewriter: PatternRewriter
    ) -> None:
        raise NotImplementedError("CeilDivUI is not supported")


lower_arith_remui = LowerBinaryIntegerOp(arith.RemUIOp, riscv.RemuOp)
lower_arith_remsi = LowerBinaryIntegerOp(arith.RemSIOp, riscv.RemOp)


class LowerArithMinSI(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.MinSIOp, rewriter: PatternRewriter) -> None:
        raise NotImplementedError("MinSI is not supported")


class LowerArithMaxSI(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.MaxSIOp, rewriter: PatternRewriter) -> None:
        raise NotImplementedError("MaxSI is not supported")


class LowerArithMinUI(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.MinUIOp, rewriter: PatternRewriter) -> None:
        raise NotImplementedError("MinUI is not supported")


class LowerArithMaxUI(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.MaxUIOp, rewriter: PatternRewriter) -> None:
        raise NotImplementedError("MaxUI is not supported")


class LowerArithCmpi(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.CmpiOp, rewriter: PatternRewriter) -> None:
        # based on https://github.com/llvm/llvm-project/blob/main/llvm/test/CodeGen/RISCV/i32-icmp.ll
        lhs, rhs = cast_operands_to_regs(rewriter)
        cast_matched_op_results(rewriter)

        match op.predicate.value.data:
            # eq
            case 0:
                xor_op = riscv.XorOp(lhs, rhs)
                seqz_op = riscv.SltiuOp(xor_op, 1)
                rewriter.replace_matched_op([xor_op, seqz_op])
            # ne
            case 1:
                zero = riscv.GetRegisterOp(riscv.Registers.ZERO)
                xor_op = riscv.XorOp(lhs, rhs)
                snez_op = riscv.SltuOp(zero, xor_op)
                rewriter.replace_matched_op([zero, xor_op, snez_op])
            # slt
            case 2:
                rewriter.replace_matched_op([riscv.SltOp(lhs, rhs)])
            # sle
            case 3:
                slt = riscv.SltOp(lhs, rhs)
                xori = riscv.XoriOp(slt, 1)
                rewriter.replace_matched_op([slt, xori])
            # ult
            case 4:
                rewriter.replace_matched_op([riscv.SltuOp(lhs, rhs)])
            # ule
            case 5:
                sltu = riscv.SltuOp(lhs, rhs)
                xori = riscv.XoriOp(sltu, 1)
                rewriter.replace_matched_op([sltu, xori])
            # ugt
            case 6:
                rewriter.replace_matched_op([riscv.SltuOp(rhs, lhs)])
            # uge
            case 7:
                sltu = riscv.SltuOp(rhs, lhs)
                xori = riscv.XoriOp(sltu, 1)
                rewriter.replace_matched_op([sltu, xori])
            case _:
                raise NotImplementedError("Cmpi predicate not supported")


class LowerArithSelect(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.SelectOp, rewriter: PatternRewriter) -> None:
        raise NotImplementedError("Select is not supported")


lower_arith_andi = LowerBinaryIntegerOp(arith.AndIOp, riscv.AndOp)
lower_arith_ori = LowerBinaryIntegerOp(arith.OrIOp, riscv.OrOp)
lower_arith_xori = LowerBinaryIntegerOp(arith.XOrIOp, riscv.XorOp)
lower_arith_shli = LowerBinaryIntegerOp(arith.ShLIOp, riscv.SllOp)
lower_arith_shrui = LowerBinaryIntegerOp(arith.ShRUIOp, riscv.SrlOp)
lower_arith_shrsi = LowerBinaryIntegerOp(arith.ShRSIOp, riscv.SraOp)


lower_arith_addf = LowerBinaryFloatOp(arith.AddfOp, riscv.FAddSOp, riscv.FAddDOp)
lower_arith_subf = LowerBinaryFloatOp(arith.SubfOp, riscv.FSubSOp, riscv.FSubDOp)
lower_arith_mulf = LowerBinaryFloatOp(arith.MulfOp, riscv.FMulSOp, riscv.FMulDOp)
lower_arith_divf = LowerBinaryFloatOp(arith.DivfOp, riscv.FDivSOp, riscv.FDivDOp)
lower_arith_minf = LowerBinaryFloatOp(arith.MinimumfOp, riscv.FMinSOp, riscv.FMinDOp)
lower_arith_maxf = LowerBinaryFloatOp(arith.MaximumfOp, riscv.FMaxSOp, riscv.FMaxDOp)


class LowerArithNegf(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.NegfOp, rewriter: PatternRewriter) -> None:
        rewriter.replace_matched_op(
            (
                operand := UnrealizedConversionCastOp.get(
                    (op.operand,), (_FLOAT_REGISTER_TYPE,)
                ),
                negf := riscv.FSgnJNSOp(operand, operand),
                UnrealizedConversionCastOp.get((negf.rd,), (op.result.type,)),
            )
        )


class LowerArithCmpf(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.CmpfOp, rewriter: PatternRewriter) -> None:
        # https://llvm.org/docs/LangRef.html#id309
        lhs, rhs = cast_operands_to_regs(rewriter)
        cast_matched_op_results(rewriter)

        fastmath = riscv.FastMathFlagsAttr(op.fastmath.data)

        match op.predicate.value.data:
            # false
            case 0:
                rewriter.replace_matched_op([riscv.LiOp(0)])
            # oeq
            case 1:
                rewriter.replace_matched_op([riscv.FeqSOp(lhs, rhs, fastmath=fastmath)])
            # ogt
            case 2:
                rewriter.replace_matched_op([riscv.FltSOp(rhs, lhs, fastmath=fastmath)])
            # oge
            case 3:
                rewriter.replace_matched_op([riscv.FleSOp(rhs, lhs, fastmath=fastmath)])
            # olt
            case 4:
                rewriter.replace_matched_op([riscv.FltSOp(lhs, rhs, fastmath=fastmath)])
            # ole
            case 5:
                rewriter.replace_matched_op([riscv.FleSOp(lhs, rhs, fastmath=fastmath)])
            # one
            case 6:
                flt1 = riscv.FltSOp(lhs, rhs, fastmath=fastmath)
                flt2 = riscv.FltSOp(rhs, lhs, fastmath=fastmath)
                rewriter.replace_matched_op(
                    [
                        flt1,
                        flt2,
                        riscv.OrOp(flt2, flt1),
                    ]
                )
            # ord
            case 7:
                feq1 = riscv.FeqSOp(lhs, lhs, fastmath=fastmath)
                feq2 = riscv.FeqSOp(rhs, rhs, fastmath=fastmath)
                rewriter.replace_matched_op(
                    [
                        feq1,
                        feq2,
                        riscv.AndOp(feq2, feq1),
                    ]
                )
            # ueq
            case 8:
                flt1 = riscv.FltSOp(lhs, rhs, fastmath=fastmath)
                flt2 = riscv.FltSOp(rhs, lhs, fastmath=fastmath)
                or_ = riscv.OrOp(flt2, flt1)
                rewriter.replace_matched_op([flt1, flt2, or_, riscv.XoriOp(or_, 1)])
            # ugt
            case 9:
                fle = riscv.FleSOp(lhs, rhs, fastmath=fastmath)
                rewriter.replace_matched_op([fle, riscv.XoriOp(fle, 1)])
            # uge
            case 10:
                fle = riscv.FltSOp(lhs, rhs, fastmath=fastmath)
                rewriter.replace_matched_op([fle, riscv.XoriOp(fle, 1)])
            # ult
            case 11:
                fle = riscv.FleSOp(rhs, lhs, fastmath=fastmath)
                rewriter.replace_matched_op([fle, riscv.XoriOp(fle, 1)])
            # ule
            case 12:
                flt = riscv.FltSOp(rhs, lhs, fastmath=fastmath)
                rewriter.replace_matched_op([flt, riscv.XoriOp(flt, 1)])
            # une
            case 13:
                feq = riscv.FeqSOp(lhs, rhs, fastmath=fastmath)
                rewriter.replace_matched_op([feq, riscv.XoriOp(feq, 1)])
            # uno
            case 14:
                feq1 = riscv.FeqSOp(lhs, lhs, fastmath=fastmath)
                feq2 = riscv.FeqSOp(rhs, rhs, fastmath=fastmath)
                and_ = riscv.AndOp(feq2, feq1)
                rewriter.replace_matched_op([feq1, feq2, and_, riscv.XoriOp(and_, 1)])
            # true
            case 15:
                rewriter.replace_matched_op([riscv.LiOp(1)])
            case _:
                raise NotImplementedError("Cmpf predicate not supported")


class LowerArithSIToFPOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.SIToFPOp, rewriter: PatternRewriter) -> None:
        match op.result.type:
            case Float32Type():
                cls = riscv.FCvtSWOp
            case Float64Type():
                cls = riscv.FCvtDWOp
            case _:
                raise ValueError(f"Unexpected float type {op.result.type}")

        rewriter.replace_matched_op(
            (
                cast_input := UnrealizedConversionCastOp.get(
                    (op.input,), (_INT_REGISTER_TYPE,)
                ),
                new_op := cls(cast_input.results[0]),
                UnrealizedConversionCastOp.get((new_op.rd,), (op.result.type,)),
            )
        )


class LowerArithFPToSIOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.FPToSIOp, rewriter: PatternRewriter) -> None:
        rewriter.replace_matched_op(
            (
                cast_input := UnrealizedConversionCastOp.get(
                    (op.input,), (_FLOAT_REGISTER_TYPE,)
                ),
                new_op := riscv.FCvtWSOp(cast_input.results[0]),
                UnrealizedConversionCastOp.get((new_op.rd,), (op.result.type,)),
            )
        )


class LowerArithExtFOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.ExtFOp, rewriter: PatternRewriter) -> None:
        raise NotImplementedError("ExtF is not supported")


class LowerArithTruncFOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.TruncFOp, rewriter: PatternRewriter) -> None:
        raise NotImplementedError("TruncF is not supported")


class ConvertArithToRiscvPass(ModulePass):
    name = "convert-arith-to-riscv"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        walker = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    LowerArithConstant(),
                    LowerArithIndexCast(),
                    LowerArithSIToFPOp(),
                    LowerArithFPToSIOp(),
                    lower_arith_addi,
                    lower_arith_subi,
                    lower_arith_muli,
                    lower_arith_divui,
                    lower_arith_divsi,
                    LowerArithFloorDivSI(),
                    lower_arith_remsi,
                    LowerArithCmpi(),
                    lower_arith_addf,
                    lower_arith_subf,
                    lower_arith_divf,
                    LowerArithNegf(),
                    lower_arith_mulf,
                    LowerArithCmpf(),
                    lower_arith_remui,
                    lower_arith_andi,
                    lower_arith_ori,
                    lower_arith_xori,
                    lower_arith_shli,
                    lower_arith_shrui,
                    lower_arith_shrsi,
                    LowerArithCeilDivSI(),
                    LowerArithCeilDivUI(),
                    LowerArithMinSI(),
                    LowerArithMaxSI(),
                    LowerArithMinUI(),
                    LowerArithMaxUI(),
                    LowerArithSelect(),
                    LowerArithExtFOp(),
                    LowerArithTruncFOp(),
                    lower_arith_minf,
                    lower_arith_maxf,
                ]
            ),
            apply_recursively=False,
        )
        walker.rewrite_module(op)



from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import NamedTuple

from xdsl.utils.lexer import Location

INDENT = 2


class VarType(NamedTuple):
    "A variable type with shape information."

    shape: list[int]


class Dumper(NamedTuple):
    lines: list[str]
    indentation: int = 0

    def append(self, prefix: str, line: str):
        self.lines.append(" " * self.indentation * INDENT + prefix + line)

    def append_list(
        self,
        prefix: str,
        open_paren: str,
        exprs: Iterable[ExprAST | FunctionAST],
        close_paren: str,
        block: Callable[[Dumper, ExprAST | FunctionAST], None],
    ):
        self.append(prefix, open_paren)
        child = self.child()
        for expr in exprs:
            block(child, expr)
        self.append("", close_paren)

    def child(self):
        return Dumper(self.lines, self.indentation + 1)

    @property
    def message(self):
        return "\n".join(self.lines)


class VarDeclExprAST(NamedTuple):
    "Expression class for defining a variable."

    loc: Location
    name: str
    varType: VarType
    expr: ExprAST

    def inner_dump(self, prefix: str, dumper: Dumper):
        dims_str = ", ".join(f"{int(dim)}" for dim in self.varType.shape)
        dumper.append("VarDecl ", f"{self.name}<{dims_str}> @{self.loc}")
        child = dumper.child()
        self.expr.inner_dump("", child)


class ReturnExprAST(NamedTuple):
    "Expression class for a return operator."

    loc: Location
    expr: ExprAST | None

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append(prefix, "Return")
        if self.expr is not None:
            child = dumper.child()
            self.expr.inner_dump("", child)


class NumberExprAST(NamedTuple):
    'Expression class for numeric literals like "1.0".'

    loc: Location
    val: float

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append(prefix, f" {self.val:.6e}")


class LiteralExprAST(NamedTuple):
    "Expression class for a literal value."

    loc: Location
    values: list[LiteralExprAST | NumberExprAST]
    dims: list[int]

    def __dump(self) -> str:
        dims_str = ", ".join(f"{int(dim)}" for dim in self.dims)
        vals_str = ",".join(
            val.__dump() if isinstance(val, LiteralExprAST) else f" {val.val:.6e}"
            for val in self.values
        )
        return f" <{dims_str}>[{vals_str}]"

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append("Literal:", self.__dump() + f" @{self.loc}")


class VariableExprAST(NamedTuple):
    'Expression class for referencing a variable, like "a".'

    loc: Location
    name: str

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append("var: ", f"{self.name} @{self.loc}")


class BinaryExprAST(NamedTuple):
    "Expression class for a binary operator."

    loc: Location
    op: str
    lhs: ExprAST
    rhs: ExprAST

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append(prefix, f"BinOp: {self.op} @{self.loc}")
        child = dumper.child()
        self.lhs.inner_dump("", child)
        self.rhs.inner_dump("", child)


class CallExprAST(NamedTuple):
    "Expression class for function calls."

    loc: Location
    callee: str
    args: list[ExprAST]

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append_list(
            prefix,
            f"Call '{self.callee}' [ @{self.loc}",
            self.args,
            "]",
            lambda dd, arg: arg.inner_dump("", dd),
        )


class PrintExprAST(NamedTuple):
    "Expression class for builtin print calls."

    loc: Location
    arg: ExprAST

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append(prefix, self.__class__.__name__)
        child = dumper.child()
        self.arg.inner_dump("arg: ", child)


class PrototypeAST(NamedTuple):
    """
    This class represents the "prototype" for a function, which captures its
    name, and its argument names (thus implicitly the number of arguments the
    function takes).
    """

    loc: Location
    name: str
    args: list[VariableExprAST]

    def dump(self):
        dumper = Dumper([])
        self.inner_dump("", dumper)
        return dumper.message

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append("", f"Proto '{self.name}' @{self.loc}")
        dumper.append("Params: ", f"[{', '.join(arg.name for arg in self.args)}]")


class FunctionAST(NamedTuple):
    "This class represents a function definition itself."

    loc: Location
    proto: PrototypeAST
    body: tuple[ExprAST, ...]

    def dump(self):
        dumper = Dumper([])
        self.inner_dump("", dumper)
        return dumper.message

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append(prefix, "Function")
        child = dumper.child()
        self.proto.inner_dump("proto: ", child)
        child.append_list(
            "Block ",
            "{",
            self.body,
            "} // Block",
            lambda dd, stmt: stmt.inner_dump("", dd),
        )


class ModuleAST(NamedTuple):
    "This class represents a list of functions to be processed together"

    funcs: tuple[FunctionAST, ...]

    def dump(self):
        dumper = Dumper([])
        self.inner_dump("", dumper)
        return dumper.message

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append_list(
            prefix, "Module:", self.funcs, "", lambda dd, func: func.inner_dump("", dd)
        )


ExprAST = (
    BinaryExprAST
    | VariableExprAST
    | LiteralExprAST
    | CallExprAST
    | NumberExprAST
    | PrintExprAST
    | VarDeclExprAST
    | ReturnExprAST
)



import keyword

from xdsl.builder import Builder, InsertPoint
from xdsl.dialects.builtin import ArrayAttr, StringAttr
from xdsl.dialects.irdl import AnyOp
from xdsl.dialects.irdl.irdl import (
    AttributeOp,
    DialectOp,
    OperandsOp,
    OperationOp,
    ParametersOp,
    ResultsOp,
    VariadicityArrayAttr,
    VariadicityAttr,
)
from xdsl.ir import Block, Dialect, ParametrizedAttribute, Region, SSAValue
from xdsl.irdl import (
    AttrConstraint,
    IRDLOperation,
    OptionalDef,
    RangeConstraint,
    VariadicDef,
)


def depython_name(name: str):
    if name[-1] == "_" and keyword.iskeyword(name[:-1]):
        return name[:-1]
    return name


def constraint_to_irdl(builder: Builder, constraint: AttrConstraint) -> SSAValue:
    """
    Convert an attribute constraint to IRDL.
    This will create new operations at the provided builder location.
    """
    any_op = builder.insert(AnyOp())
    return any_op.output


def range_to_irdl(builder: Builder, constraint: RangeConstraint) -> SSAValue:
    """
    Convert a range constraint to IRDL.
    This will create new operations at the provided builder location.
    """
    any_op = builder.insert(AnyOp())
    return any_op.output


def op_def_to_irdl(op: type[IRDLOperation]) -> OperationOp:
    """Convert an operation definition to an IRDL operation definition."""
    op_def = op.get_irdl_definition()

    block = Block()
    builder = Builder(InsertPoint.at_end(block))

    # Operands
    operand_values: list[SSAValue] = []
    operand_variadicities: list[VariadicityAttr] = []
    operand_names: list[StringAttr] = []
    for operand in op_def.operands:
        operand_values.append(range_to_irdl(builder, operand[1].constr))
        if isinstance(operand[1], OptionalDef):
            operand_variadicities.append(VariadicityAttr.OPTIONAL)
        elif isinstance(operand[1], VariadicDef):
            operand_variadicities.append(VariadicityAttr.VARIADIC)
        else:
            operand_variadicities.append(VariadicityAttr.SINGLE)
        operand_names.append(StringAttr(depython_name(operand[0])))
    if operand_values:
        builder.insert(
            OperandsOp(
                operand_values,
                VariadicityArrayAttr(ArrayAttr(operand_variadicities)),
                ArrayAttr(operand_names),
            )
        )

    # Results
    result_values: list[SSAValue] = []
    result_variadicities: list[VariadicityAttr] = []
    result_names: list[StringAttr] = []
    for result in op_def.results:
        result_values.append(range_to_irdl(builder, result[1].constr))
        if isinstance(result[1], OptionalDef):
            result_variadicities.append(VariadicityAttr.OPTIONAL)
        elif isinstance(result[1], VariadicDef):
            result_variadicities.append(VariadicityAttr.VARIADIC)
        else:
            result_variadicities.append(VariadicityAttr.SINGLE)
        result_names.append(StringAttr(depython_name(result[0])))
    if result_values:
        builder.insert(
            ResultsOp(
                result_values,
                VariadicityArrayAttr(ArrayAttr(result_variadicities)),
                ArrayAttr(result_names),
            )
        )

    return OperationOp(Dialect.split_name(op_def.name)[1], Region([block]))


def attr_def_to_irdl(
    attr: type[ParametrizedAttribute],
) -> AttributeOp:
    """Convert an attribute definition to an IRDL attribute definition."""
    attr_def = attr.get_irdl_definition()

    block = Block()
    builder = Builder(InsertPoint.at_end(block))

    # Parameters
    param_values: list[SSAValue] = []
    names: list[StringAttr] = []
    for param in attr_def.parameters:
        param_values.append(constraint_to_irdl(builder, param[1].constr))
        names.append(StringAttr(depython_name(param[0])))
    builder.insert(ParametersOp(param_values, ArrayAttr(names)))

    return AttributeOp(Dialect.split_name(attr_def.name)[1], Region([block]))


def dialect_to_irdl(dialect: Dialect, name: str) -> DialectOp:
    """Convert a dialect definition to an IRDL dialect definition."""
    block = Block()
    builder = Builder(InsertPoint.at_end(block))

    for attribute in dialect.attributes:
        if not issubclass(attribute, ParametrizedAttribute):
            raise ValueError(
                "Can only convert ParametrizedAttribute attributes to IRDL"
            )
        builder.insert(attr_def_to_irdl(attribute))

    for operation in dialect.operations:
        if not issubclass(operation, IRDLOperation):
            raise ValueError("Can only convert IRDLOperations operations to IRDL")
        builder.insert(op_def_to_irdl(operation))

    return DialectOp(name, Region([block]))



import json
from collections import defaultdict
from collections.abc import Iterable

from xdsl.backend.block_naive_allocator import BlockNaiveAllocator
from xdsl.backend.register_allocatable import RegisterAllocatableOperation
from xdsl.backend.register_allocator import live_ins_per_block
from xdsl.backend.register_stack import RegisterStack
from xdsl.backend.register_type import RegisterType
from xdsl.dialects import riscv, riscv_func
from xdsl.dialects.riscv import Registers, RISCVRegisterType
from xdsl.ir import SSAValue
from xdsl.rewriter import InsertPoint, Rewriter
from xdsl.transforms.canonicalization_patterns.riscv import get_constant_value


def reg_types_by_name(regs: Iterable[RISCVRegisterType]) -> dict[str, set[str]]:
    """
    Groups register types by name.
    """
    res = defaultdict[str, set[str]](set)
    for reg in regs:
        res[reg.name].add(reg.register_name.data)
    return res


class RegisterAllocatorLivenessBlockNaive(BlockNaiveAllocator):
    def __init__(self, available_registers: RegisterStack) -> None:
        super().__init__(available_registers, RISCVRegisterType)

    def new_type_for_value(self, reg: SSAValue) -> RegisterType | None:
        if (
            isinstance(reg.type, self.register_base_class)
            and not reg.type.is_allocated
            and (val := get_constant_value(reg)) is not None
            and val.value.data == 0
        ):
            return Registers.ZERO
        return super().new_type_for_value(reg)

    def allocate_func(
        self, func: riscv_func.FuncOp, *, add_regalloc_stats: bool = False
    ) -> None:
        """
        Allocates values in function passed in to registers.
        The whole function must have been lowered to the relevant riscv dialects
        and it must contain no unrealized casts.
        If `add_regalloc_stats` is set to `True`, then a comment op will be inserted
        before the function op passed in with a json containing the relevant data.
        """
        if not func.body.blocks:
            # External function declaration
            return

        if len(func.body.blocks) != 1:
            raise NotImplementedError(
                f"Cannot register allocate func with {len(func.body.blocks)} blocks."
            )

        preallocated = {
            reg
            for reg in RegisterAllocatableOperation.iter_all_used_registers(func.body)
            if isinstance(reg, RISCVRegisterType)
        }

        for pa_reg in preallocated:
            self.available_registers.exclude_register(pa_reg)

        block = func.body.block

        self.live_ins_per_block = live_ins_per_block(block)
        assert not self.live_ins_per_block[block]

        self.allocate_block(block)

        if add_regalloc_stats:
            preallocated_stats = reg_types_by_name(preallocated)
            allocated_stats = reg_types_by_name(
                val.type
                for op in block.walk()
                for vals in (op.results, op.operands)
                for val in vals
                if isinstance(val.type, RISCVRegisterType)
            )
            stats = {
                "preallocated_float": sorted(preallocated_stats["riscv.freg"]),
                "preallocated_int": sorted(preallocated_stats["riscv.reg"]),
                "allocated_float": sorted(allocated_stats["riscv.freg"]),
                "allocated_int": sorted(allocated_stats["riscv.reg"]),
            }

            stats_str = json.dumps(stats)

            Rewriter.insert_op(
                riscv.CommentOp(f"Regalloc stats: {stats_str}"),
                InsertPoint.before(func),
            )



from io import StringIO

from xdsl.dialects import riscv, riscv_debug
from xdsl.dialects.builtin import ModuleOp
from xdsl.interpreter import Interpreter
from xdsl.interpreters.riscv import RiscvFunctions
from xdsl.interpreters.riscv_debug import RiscvDebugFunctions
from xdsl.utils.test_value import create_ssa_value


def test_riscv_interpreter():
    module_op = ModuleOp([])

    riscv_functions = RiscvFunctions()
    riscv_debug_functions = RiscvDebugFunctions()
    file = StringIO()
    interpreter = Interpreter(module_op, file=file)
    interpreter.register_implementations(riscv_functions)
    interpreter.register_implementations(riscv_debug_functions)

    assert (
        interpreter.run_op(
            riscv_debug.PrintfOp(
                "{} {} {} {}",
                (
                    create_ssa_value(riscv.Registers.UNALLOCATED_INT),
                    create_ssa_value(riscv.Registers.UNALLOCATED_INT),
                    create_ssa_value(riscv.Registers.UNALLOCATED_FLOAT),
                    create_ssa_value(riscv.Registers.UNALLOCATED_FLOAT),
                ),
            ),
            (1, -1, 2.0, -2.0),
        )
        == ()
    )

    assert file.getvalue() == "1 -1 2.0 -2.0"



from dataclasses import dataclass

from xdsl.context import Context
from xdsl.dialects import builtin, symref
from xdsl.frontend.pyast.utils.exceptions import FrontendProgramException
from xdsl.ir import Block, Operation, Region
from xdsl.passes import ModulePass
from xdsl.rewriter import Rewriter

# Background
# ==========
#
# We want to allow users to write non-SSA code. For example, the frontend should
# allow functions like this:
#
# ```
# def foo() -> i32:
#   a: i32 = 0
#   for i in range(100):
#     a = a + 2
#   return a
# ```
#
# The solution is to have a `symref` dialect: it uses declare, fetch and store
# operations (for those familiar with LLVM, these correspond to `alloca`, `load`
# and `store` but based on symbols instead of memory locations). Each symref
# operation is defined by the programmer with a simple mapping:
#
# 1. variable declaration, e.g. `a: i32 = 0` maps to `declare @a`
# 2. variable use, e.g. `... = a` maps to `... = fetch @a`
# 3. variable assignement, e.g. `a = ...` maps to `update @a ...`
#
# With these, it is relatively straightforward to lower any Python program to
# xDSL (taking care of scoping rules, of course). For example, the code above
# would become (in a slightly abused xDSL syntax):
#
# ```
# func foo() -> i32 {
#   declare @a
#   update @a with 0
#   for 0 to 100 {
#     t1 = fetch @a
#     t2 = add t1 2
#     update @a t2
#   }
#   t3 = fetch @a
#   return t3
# }
# ```
#
# Note that while insertion of symref operations is easy, the generated xDSL is
# not fully in SSA form. For example, blocks in region should pass values
# between each other via block arguments, instead of symref calls. Similarly,
# some operations like scf.if or affine.for should yield a value.
#
#
# Desymrefication pass
# ====================
#
# In this file, we implement desymrefication pass - it goes over generated xDSL
# and removes all symref operations. To describe how the pass works, first
# recall that in xDSL an operation can contain regions which in turn can contain
# CFGs of basic blocks. An operation passes control flow to a region and on exit
# from the region the control flow returns to the operation. Then the operation
# can transfer control flow to another region, etc.
#
# Desymrefication can be applied to regions and operations. First, consider a
# simple case when all operations in the region do not have any nested regions.
# This means that the region is simply a CFG. Otherwise, there are some
# operations that contain regions. However, we can apply desymrefication to
# these operations them first, reducing the problem to the the first case. In
# pseudocode, that would look like:
#
# ```
# desymref_region(region):
#   for block in region.blocks:
#     for op in block.ops:
#       desymref_op(op)
#     # at this point no op contains a symref operation
#     # in the nested region.
# ```
#
# Next, we describe how desymrefication actually works, on a high level.
# Every region either declares a symbol or uses it from the parent region.
# Let's consider each case separately.
#
# Case 1: Symbol is declared in this region. This is an easy case. We already
# know that symbol is only used in this region and no other nested region uses
# it. Therefore running any SSA-construction algorithm on the CFG is enough.
#
# Case 2: Symbol is not declared in this region. In this case, we only support
# non-CFGs at the moment (i.e. single block regions). We observe that in general
# every symbol can be simplified within a block to have at most one fetch (start
# of the block) at most one update (end of the block). This means that using
# specification of an operation we can promote these symbols outside.
# TODO: Add op promotion.


def has_symbol(op: Operation) -> bool:
    return isinstance(op, symref.DeclareOp | symref.FetchOp | symref.UpdateOp)


def get_symbol(op: Operation) -> str | None:
    """
    Returns a symbol attribute for an operation. If operation has no symbol
    attributes, None is returned.
    """
    if isinstance(op, symref.DeclareOp):
        return op.sym_name.data
    elif isinstance(op, symref.FetchOp | symref.UpdateOp):
        return op.symbol.root_reference.data
    else:
        return None


def get_symbols(block: Block) -> set[str]:
    """Returns a set of all symbols defined/used in a basic block."""
    symbols: set[str] = set()
    for op in block.ops:
        if has_symbol(op):
            symbol = get_symbol(op)
            assert symbol is not None
            symbols.add(symbol)
    return symbols


def lower_positional_bound(
    writes: list[symref.UpdateOp], read: symref.FetchOp
) -> Operation | None:
    """
    Returns a nearest write preceeding the `read`. If there is no such write,
    `None` is returned.

    Pre-condition: list `writes` is sorted based on operation indices.
    """
    block = read.parent_block()
    assert block is not None

    idx = block.get_operation_index(read)
    low_idx = -1
    high_idx = len(writes) - 1

    # Binary search to find the right write.
    while low_idx < high_idx:
        mid_idx = (high_idx - low_idx + 1) // 2 + low_idx
        user_idx = block.get_operation_index(writes[mid_idx])

        if user_idx < idx:
            low_idx = mid_idx
        else:
            high_idx = mid_idx - 1

    if low_idx == -1:
        return None
    return writes[low_idx]


@dataclass
class Desymrefier:
    """
    Rewrites the program by removing all reads/writes from/to symbols and symbol
    definitions.
    """

    def desymrefy(self, op: Operation):
        """
        Desymrefy an operation. This method guarantees that the operation does
        not have any symbols.
        """
        self.prepare_op(op)
        self.promote_op(op)

    def promote_op(self, op: Operation):
        """
        Promotes an operation. This method guarantees that the operation does
        not have any symbols.
        """
        # TODO: Add promoters in the next patch.
        pass

    def prepare_op(self, op: Operation):
        """
        Prepares an operation for promotion. This method guarantees that any
        symbol in any region of this operation is read at most once and written
        at most once.
        """

        # For operation with no regions we don't have to do any work.
        if len(op.regions) == 0:
            return

        # Otherwise, we have to prepare regions.
        for region in op.regions:
            self.prepare_region(region)

    def prepare_region(self, region: Region):
        """
        Prepares a region for promotion. This method guarantees that any symbol
        in the region is read at most once and written at most once.
        """
        num_blocks = len(region.blocks)
        if num_blocks == 1:
            # If there is only one block, preparing region is easier, so we
            # handle it separately.
            self.prepare_block(region.block)
        else:
            # TODO: Support regions with multiple blocks. This is not too
            # difficult but requires many more things:
            # 1. SSA-construction algorithm to prune symbol declarations. This
            #    also requires analyses passes (dominators).
            # 2. Insertion of entry/exit blocks to ensure dominance.
            raise FrontendProgramException(
                f"Running desymrefier on region with {num_blocks} > 1 blocks is "
                "not supported."
            )

    def prepare_block(self, block: Block):
        """Prepares a block for promotion."""

        # First, desymrefy nested regions.
        for op in block.ops:
            self.desymrefy(op)

        self.prune_definitions(block)
        self.prune_uses_without_definitions(block)

        # Ensure that each symbol is read/written at most once.
        symbols = get_symbols(block)
        for symbol in symbols:
            num_reads = sum(
                isinstance(op, symref.FetchOp)
                for op in block.ops
                if get_symbol(op) == symbol
            )
            num_writes = sum(
                isinstance(op, symref.UpdateOp)
                for op in block.ops
                if get_symbol(op) == symbol
            )
            if num_reads > 1 or num_writes > 1:
                raise FrontendProgramException(
                    f"Block {block} not ready for promotion: found {num_reads}"
                    f" reads and {num_writes} writes."
                )

    def prune_definitions(self, block: Block):
        """Removes all symbol definitions and their uses from the block."""

        # Find all symbol definitions in this block. If no definitions
        # found, terminate.
        while (
            len(
                definitions := [
                    op for op in block.ops if isinstance(op, symref.DeclareOp)
                ]
            )
            > 0
        ):
            # Otherwise, some definitions are still alive.
            for definition in definitions:
                symbol = get_symbol(definition)

                # Find all reads and writes for this symbol.
                reads = [
                    op
                    for op in block.ops
                    if isinstance(op, symref.FetchOp) and get_symbol(op) == symbol
                ]
                writes = [
                    op
                    for op in block.ops
                    if isinstance(op, symref.UpdateOp) and get_symbol(op) == symbol
                ]

                # Symbol is never read, so remove its definition and any writes.
                if len(reads) == 0:
                    for write in writes:
                        Rewriter.erase_op(write)
                    Rewriter.erase_op(definition)
                    continue

                # For symbols which are written once, the write dominates all
                # the uses and therefore can be trivially replaced.
                if len(writes) == 1:
                    write = writes[0]
                    for read in reads:
                        Rewriter.replace_op(read, [], [write.operands[0]])
                    Rewriter.erase_op(write)
                    Rewriter.erase_op(definition)
                    continue

                # If there are multiple reads and writes, replace every
                # read with the closest preceding write.
                for read in reads:
                    write = lower_positional_bound(writes, read)
                    if write is not None:
                        Rewriter.replace_op(read, [], [write.operands[0]])

    def _prune_unused_reads(self, block: Block):
        def is_unused_read(op: Operation) -> bool:
            return isinstance(op, symref.FetchOp) and not op.results[0].uses

        unused_reads = [op for op in block.ops if is_unused_read(op)]
        for read in unused_reads:
            Rewriter.erase_op(read)

    def prune_uses_without_definitions(self, block: Block):
        """Removes all possible symbol uses in a single block."""
        prepared_symbols: set[str] = set()

        while True:
            self._prune_unused_reads(block)

            # Find all symbols that are still in use in this block.
            symbol_worklist: set[str] = {
                symbol
                for symbol in get_symbols(block)
                if symbol not in prepared_symbols
            }
            if len(symbol_worklist) == 0:
                return

            for symbol in symbol_worklist:
                reads = [
                    op
                    for op in block.ops
                    if isinstance(op, symref.FetchOp) and get_symbol(op) == symbol
                ]
                writes = [
                    op
                    for op in block.ops
                    if isinstance(op, symref.UpdateOp) and get_symbol(op) == symbol
                ]
                assert len(reads) > 0 or len(writes) > 0

                # There are no reads, so we can only keep the last write to the
                # symbol.
                if len(reads) == 0:
                    for write in writes[:-1]:
                        Rewriter.erase_op(write)
                    prepared_symbols.add(symbol)
                    continue

                # There are no writes, so we can replace all reads with this
                # symbol.
                if len(writes) == 0:
                    for read in reads[1:]:
                        Rewriter.replace_op(read, [], [reads[0].results[0]])
                    prepared_symbols.add(symbol)
                    continue

                # sets of reads and writes are disjoint.
                last_read_idx = block.get_operation_index(reads[-1])
                first_write_idx = block.get_operation_index(writes[0])
                if last_read_idx < first_write_idx:
                    for read in reads[1:]:
                        Rewriter.replace_op(read, [], [reads[0].results[0]])
                    for write in writes[:-1]:
                        Rewriter.erase_op(write)
                    prepared_symbols.add(symbol)
                    continue

                # Otherwise, replace reads with the closest preceding write.
                for read in reads:
                    write = lower_positional_bound(writes, read)
                    if write is not None:
                        Rewriter.replace_op(read, [], [write.operands[0]])


class FrontendDesymrefyPass(ModulePass):
    name = "frontend-desymrefy"

    def apply(self, ctx: Context, op: builtin.ModuleOp):
        Desymrefier().desymrefy(op)



from xdsl.dialects.builtin import ArrayAttr, IntAttr, StringAttr


def test_data_get():
    s = "Hello world"
    s_attr = StringAttr(s)
    assert StringAttr.get(s) == s_attr
    assert StringAttr.get(s_attr) == s_attr

    i = 42
    i_attr = IntAttr(i)
    assert IntAttr.get(i) == i_attr
    assert IntAttr.get(i_attr) == i_attr

    a = (
        IntAttr(1),
        IntAttr(2),
        IntAttr(3),
    )
    a_attr = ArrayAttr(a)
    assert ArrayAttr.get(a) == a_attr
    assert ArrayAttr.get(a_attr) == a_attr



from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import arith, func, math, scf
from xdsl.dialects.builtin import IndexType, IntegerType, ModuleOp, i32
from xdsl.dialects.printf import PrintCharOp, PrintIntOp
from xdsl.ir import Block, SSAValue
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.traits import SymbolTable

i8 = IntegerType(8)


class LowerPrintCharToPutchar(RewritePattern):
    """
    Rewrite Pattern that rewrites printf.print_char to an
    external function call
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: PrintCharOp, rewriter: PatternRewriter, /):
        func_call = func.CallOp("putchar", [op.char], [i32])
        cast_op = arith.ExtUIOp(op.char, i32)
        func_call = func.CallOp("putchar", [cast_op], [i32])
        # Add empty new_results, since a result is necessary for linking
        # putchar, but the result does not exist anywhere.
        rewriter.replace_matched_op([cast_op, func_call], new_results=[])


class ConvertPrintIntToItoa(RewritePattern):
    """
    Rewrite Pattern that rewrites printf.print_int to an
    mlir_itoa function
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: PrintIntOp, rewriter: PatternRewriter, /):
        func_call = func.CallOp("mlir_itoa", [op.int], [])
        rewriter.replace_matched_op(func_call)


def get_mlir_itoa():
    """
    This function returns an MLIR func.func that holds a routine that calls
    an external putchar implementation several times on all individual digits
    that make up the integer.

    Roughly it exists of the following steps:
        1. If the number is zero, print zero and exit
        2. Check if the number is negative: if so, print the minus sign
        3. Get the number of digits for the absolute value
        4. Go through the digits in reverse order and print them one by one

    Several things to note about this specfic implementation:
        * It does not require floating point arithmetic
        * It does not require a printf implementation, only putchar.
    """

    # get_number_of_digits implementation
    def get_number_of_digits(absolute_value: SSAValue) -> SSAValue:
        """
        Return the result of an MLIR scf.while op that gets the number of
        digits of a positive integer. Roughly goes like this in
        python pseudocode (e.g. integer is 420, routine should return 3):
        ```
        integer = 420
        digits = 0
        while(integer != 0):
            integer //= 10
            digits += 1
        return digits

        ```
        """
        before_block = Block(arg_types=(i32, i32))
        after_block = Block(arg_types=(i32, i32))
        digit_init = arith.ConstantOp.from_int_and_width(0, i32)
        one = arith.ConstantOp.from_int_and_width(1, i32)
        while_loop = scf.WhileOp(
            [absolute_value, digit_init],
            [i32, i32],
            [before_block],
            [after_block],
        )
        with ImplicitBuilder(before_block) as (running_integer, digits):
            # Stop when you reach zero
            is_zero = arith.CmpiOp(running_integer, zero, "ne")
            scf.ConditionOp(is_zero, running_integer, digits)
        with ImplicitBuilder(after_block) as (running_integer, previous_digits):
            ten = arith.ConstantOp.from_int_and_width(10, 32)
            new_integer = arith.DivUIOp(running_integer, ten)
            digits = arith.AddiOp(previous_digits, one)
            scf.YieldOp(new_integer, digits)

        return while_loop.results[1]

    # print_minus_if_negative implementation
    def print_minus_if_negative(integer: SSAValue) -> SSAValue:
        """
        This function returns an MLIR scf.if op that prints a minus sign
        if the integer value is negative and returns the absolute value
        of the integer in either case, in python code:
        ```
        if(integer > 0):
            print("-")
            return (0 - integer)
        else:
            return integer
        ```
        """
        is_negative_block = Block()
        is_positive_block = Block()
        is_negative = arith.CmpiOp(integer, zero, "slt")

        # Either way get the absolute value of the number
        absolute_value = scf.IfOp(
            is_negative, [i32], [is_negative_block], [is_positive_block]
        )
        # If the value is negative, print the minus sign, return abs value
        with ImplicitBuilder(is_negative_block):
            PrintCharOp.from_constant_char("-")
            negative_input = arith.SubiOp(zero, integer)
            scf.YieldOp(negative_input)
        # If the value is positive, just return value itself as abs value
        with ImplicitBuilder(is_positive_block):
            scf.YieldOp(integer)

        return absolute_value.results[0]

    def print_digits(digits: SSAValue, absolute_value: SSAValue) -> scf.ForOp:
        """
        Return an scf.for loop that prints all digits of the given value.
        In python code:
        for i in range(0, digits, 1):
            position = (digits - 1) - i
            digit = (absolute_value // (10**position)) % 10
            print(digit)
        """
        zero_index = arith.ConstantOp.from_int_and_width(0, IndexType())
        one_index = arith.ConstantOp.from_int_and_width(1, IndexType())
        digits_index = arith.IndexCastOp(digits, IndexType())
        loop_body = Block(arg_types=(IndexType(),))

        # Print all from most significant to least
        for_loop = scf.ForOp(zero_index, digits_index, one_index, (), loop_body)
        with ImplicitBuilder(loop_body) as (index_var,):
            one = arith.ConstantOp.from_int_and_width(1, i32)
            size_minus_one = arith.SubiOp(digits, one)
            index_var_int = arith.IndexCastOp(index_var, i32)
            position = arith.SubiOp(size_minus_one, index_var_int)
            # digit = (num // (10**pos)) % 10
            ten = arith.ConstantOp.from_int_and_width(10, i32)
            i_0 = math.IPowIOp(ten, position)
            i_1 = arith.DivUIOp(absolute_value, i_0)
            digit = arith.RemUIOp(i_1, ten)
            ascii_offset = arith.ConstantOp.from_int_and_width(ord("0"), i32)
            char = arith.AddiOp(digit, ascii_offset)
            char_i8 = arith.TruncIOp(char, i8)
            PrintCharOp(char_i8)
            scf.YieldOp()
        return for_loop

    # Beginning of mlir_itoa declaration
    """
    Python code for this implementation:
    ```
    if(integer = 0):
        print("0")
    else:
        absolute_value = print_minus_if_negative(integer)
        digits = get_number_of_digits(absolute_value)
        print_digits(digits, absolute_value)
    ```
    """
    mlir_itoa_func = func.FuncOp("mlir_itoa", ((i32,), ()))
    with ImplicitBuilder(mlir_itoa_func.body) as (integer,):
        zero = arith.ConstantOp.from_int_and_width(0, i32)
        is_zero_block = Block()
        is_not_zero_block = Block()
        is_zero = arith.CmpiOp(integer, zero, "eq")
        scf.IfOp(
            is_zero,
            [],
            [is_zero_block],
            [is_not_zero_block],
        )
        # If the value is zero, just print one zero
        with ImplicitBuilder(is_zero_block):
            PrintCharOp.from_constant_char("0")
            scf.YieldOp()
        # Otherwise continue with itoa
        with ImplicitBuilder(is_not_zero_block):
            # Print minus sign if negative and return absolute value
            absolute_value = print_minus_if_negative(integer)
            # Get amount  of digits to be printed
            digits = get_number_of_digits(absolute_value)
            print_digits(digits, absolute_value)
            # Yield from is_not_zero_block
            scf.YieldOp()
        # Return from itoa function
        func.ReturnOp()

    return mlir_itoa_func


class PrintfToPutcharPass(ModulePass):
    name = "printf-to-putchar"

    # lower to func.call
    def apply(self, ctx: Context, op: ModuleOp) -> None:
        # Check if there are any printints
        contains_printint = any(
            isinstance(op_in_module, PrintIntOp) for op_in_module in op.walk()
        )
        if contains_printint:
            print_int_walker = PatternRewriteWalker(
                ConvertPrintIntToItoa(),
                apply_recursively=False,
            )
            print_int_walker.rewrite_module(op)
            # This has to happen first, since mlir_itoa
            # contains some references to printf.print_char
            # Add function implementation for mlir_itoa if it isn't there already
            SymbolTable.insert_or_update(op, get_mlir_itoa())
        # Check if there are any printchars
        contains_printchar = any(
            isinstance(op_in_module, PrintCharOp) for op_in_module in op.walk()
        )
        if contains_printchar:
            print_char_walker = PatternRewriteWalker(
                LowerPrintCharToPutchar(),
                apply_recursively=False,
            )
            print_char_walker.rewrite_module(op)
            # Add external putchar reference if not already there
            func_op = func.FuncOp.external("putchar", [i32], [i32])
            SymbolTable.insert_or_update(op, func_op)



from __future__ import annotations

import dataclasses
from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import Field, dataclass, field
from types import NoneType, UnionType
from typing import (
    Any,
    ClassVar,
    NamedTuple,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from typing_extensions import Self, TypeVar

from xdsl.context import Context
from xdsl.dialects import builtin
from xdsl.utils.hints import isa, type_repr
from xdsl.utils.parse_pipeline import (
    PassArgElementType,
    PassArgListType,
    PipelinePassSpec,
    parse_pipeline,
)

ModulePassT = TypeVar("ModulePassT", bound="ModulePass")


@dataclass(frozen=True)
class ModulePass(ABC):
    """
    A Pass is a named rewrite pass over an IR module that can accept arguments.

    All passes are expected to leave the IR in a valid state *after* application,
    meaning that a call to .verify() succeeds on the whole module. In turn, all
    passes can expect that the IR they are applied to is in a valid state. It
    is not required that the IR verifies at any point while the pass is being
    applied.

    In order to make a pass accept arguments, it must be a dataclass. Furthermore,
    only the following types are supported as argument types:

    Base types:                int | float | bool | string
    N-tuples of base types:
        tuple[int, ...], tuple[int|float, ...], tuple[int, ...] | tuple[float, ...]
    Top-level optional:        ... | None

    Pass arguments on the CLI are formatted as follows:

    CLI arg                             Mapped to class field
    -------------------------           ------------------------------
    my-pass{arg-1=1}                    arg_1: int             = 1
    my-pass{arg-1}                      arg_1: int | None      = None
    my-pass{arg-1=1,2,3}                arg_1: tuple[int, ...] = (1, 2, 3)
    my-pass{arg-1=true}                 arg_1: bool | None     = True
    """

    name: ClassVar[str]

    @abstractmethod
    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None: ...

    @classmethod
    def from_pass_spec(cls, spec: PipelinePassSpec) -> Self:
        """
        This method takes a PipelinePassSpec, does type checking on the
        arguments, and then instantiates an instance of the ModulePass
        from the spec.
        """
        if spec.name != cls.name:
            raise ValueError(
                f"Cannot create Pass {cls.name} from pass arguments for pass {spec.name}"
            )

        # normalize spec arg names:
        spec_arguments_dict: dict[str, PassArgListType] = (
            spec.normalize_arg_names().args
        )

        # get all dataclass fields
        fields: tuple[Field[Any], ...] = dataclasses.fields(cls)

        # start constructing the argument dict for the dataclass
        arg_dict = dict[str, PassArgListType | PassArgElementType | None]()

        required_fields = cls.required_fields()

        field_types = get_type_hints(cls)

        # iterate over all fields of the dataclass
        for op_field in fields:
            # ignore the name field and everything that's not used by __init__
            if op_field.name == "name" or not op_field.init:
                continue
            # check that non-optional fields are present
            if op_field.name not in spec_arguments_dict:
                if op_field.name not in required_fields:
                    arg_dict[op_field.name] = _get_default(op_field)
                    continue
                raise ValueError(f'Pass {cls.name} requires argument "{op_field.name}"')

            # convert pass arg to the correct type:
            field_type = field_types[op_field.name]
            arg_dict[op_field.name] = _convert_pass_arg_to_type(
                spec_arguments_dict.pop(op_field.name),
                field_type,
            )
            # we use .pop here to also remove the arg from the dict

        # if not all args were removed we raise an error
        if len(spec_arguments_dict) != 0:
            arguments_str = ", ".join(f'"{arg}"' for arg in spec_arguments_dict)
            fields_str = ", ".join(f'"{field.name}"' for field in fields)
            raise ValueError(
                f"Provided arguments [{arguments_str}] not found in expected pass "
                f"arguments [{fields_str}]"
            )

        # instantiate the dataclass using kwargs
        return cls(**arg_dict)

    @classmethod
    def required_fields(cls) -> set[str]:
        """
        Inspects the definition of the pass for fields that do not have default values.
        """
        return {
            field.name for field in dataclasses.fields(cls) if not _is_optional(field)
        }

    def pipeline_pass_spec(self, *, include_default: bool = False) -> PipelinePassSpec:
        """
        This function takes a ModulePass and returns a PipelinePassSpec.

        If `include_default` is `True`, then optional arguments are not included in the
        spec.
        """
        # get all dataclass fields
        fields = dataclasses.fields(self)
        args: dict[str, PassArgListType] = {}

        # iterate over all fields of the dataclass
        for op_field in fields:
            name = op_field.name
            # ignore the name field and everything that's not used by __init__
            if name == "name" or not op_field.init:
                continue

            val = getattr(self, name)

            if _is_optional(op_field):
                if val == _get_default(op_field) and not include_default:
                    continue

            if val is None:
                arg_list = ()
            elif isinstance(val, PassArgElementType):
                arg_list = (val,)
            else:
                arg_list = val

            args[name] = arg_list
        return PipelinePassSpec(self.name, args)


class PassOptionInfo(NamedTuple):
    """The name, expected type, and default value for one option of a module pass."""

    name: str
    expected_type: str
    default_value: str | None = None


def get_pass_option_infos(
    arg: type[ModulePassT],
) -> tuple[PassOptionInfo, ...]:
    """
    Returns the expected argument names, types, and optional expected values for options
    for the given pass.
    """

    return tuple(
        PassOptionInfo(
            field.name,
            type_repr(field.type),
            str(getattr(arg, field.name)).lower() if hasattr(arg, field.name) else None,
        )
        for field in dataclasses.fields(arg)
    )


@dataclass(frozen=True)
class PassPipeline:
    """
    A representation of a pass pipeline, with an optional callback to be executed
    between each of the passes.
    """

    passes: tuple[ModulePass, ...]
    """
    These will be executed sequentially during the execution of the pipeline.
    """
    callback: Callable[[ModulePass, builtin.ModuleOp, ModulePass], None] | None = field(
        default=None
    )
    """
    Function called in between every pass, taking the pass that just ran, the module,
    and the next pass.
    """

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        if not self.passes:
            # Early exit to avoid fetching a non-existing last pass.
            return
        callback = self.callback

        for prev, next in zip(self.passes[:-1], self.passes[1:]):
            prev.apply(ctx, op)
            if callback is not None:
                callback(prev, op, next)

        self.passes[-1].apply(ctx, op)

    @staticmethod
    def parse_spec(
        available_passes: dict[str, Callable[[], type[ModulePass]]],
        spec: str,
        callback: Callable[[ModulePass, builtin.ModuleOp, ModulePass], None]
        | None = None,
    ) -> PassPipeline:
        specs = tuple(parse_pipeline(spec))
        unrecognised_passes = tuple(
            p.name for p in specs if p.name not in available_passes
        )
        if unrecognised_passes:
            raise ValueError(f"Unrecognized passes: {list(unrecognised_passes)}")

        passes = tuple(available_passes[p.name]().from_pass_spec(p) for p in specs)

        return PassPipeline(passes, callback)


def _convert_pass_arg_to_type(
    value: PassArgListType, dest_type: Any
) -> PassArgListType | PassArgElementType | None:
    """
    Takes in a list of pass args, and converts them to the required type.

    value,      dest_type,      result
    []          int | None      None
    [1]         int | None      1
    [1]         tuple[int, ...] (1,)
    [1,2]       tuple[int, ...] (1,2)
    [1,2]       int | None      Error
    []          int             Error

    And so on
    """
    origin = get_origin(dest_type)

    # we need to special case optionals as [] means no option given
    if origin in [Union, UnionType]:
        if len(value) == 0:
            if NoneType in get_args(dest_type):
                return None
            else:
                raise ValueError("Argument must contain a value")

    # first check if an individual value passes the type check
    if len(value) == 1 and isa(value[0], dest_type):
        return value[0]

    # then check if n-tuple value is okay
    if isa(value, dest_type):
        return value

    # at this point we exhausted all possibilities
    raise ValueError(f"Incompatible types: given {value}, expected {dest_type}")


def _is_optional(field: Field[Any]):
    """
    Shorthand to check if the given type allows "None" as a value.
    """
    can_be_none = get_origin(field.type) in [Union, UnionType] and NoneType in get_args(
        field.type
    )
    has_default_val = field.default is not dataclasses.MISSING
    has_default_factory = field.default_factory is not dataclasses.MISSING

    return can_be_none or has_default_val or has_default_factory


def _get_default(field: Field[Any]) -> Any:
    if field.default is not dataclasses.MISSING:
        return field.default
    if field.default_factory is not dataclasses.MISSING:
        return field.default_factory()
    return None



from xdsl.backend.riscv.lowering.utils import (
    cast_block_args_to_regs,
    cast_matched_op_results,
    cast_operands_to_regs,
    move_to_unallocated_regs,
)
from xdsl.context import Context
from xdsl.dialects import builtin, riscv_scf, scf
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)


class ScfForLowering(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: scf.ForOp, rewriter: PatternRewriter) -> None:
        lb, ub, step, *args = cast_operands_to_regs(rewriter)
        new_region = rewriter.move_region_contents_to_new_regions(op.body)
        cast_block_args_to_regs(new_region.block, rewriter)
        mv_ops, values = move_to_unallocated_regs(args, op.iter_args.types)
        rewriter.insert_op_before_matched_op(mv_ops)
        cast_matched_op_results(rewriter)
        new_op = riscv_scf.ForOp(lb, ub, step, values, new_region)
        mv_res_ops, res_values = move_to_unallocated_regs(
            new_op.results, op.iter_args.types
        )

        rewriter.replace_matched_op((new_op, *mv_res_ops), res_values)


class ScfYieldLowering(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: scf.YieldOp, rewriter: PatternRewriter) -> None:
        rewriter.replace_matched_op(riscv_scf.YieldOp(*cast_operands_to_regs(rewriter)))


class ConvertScfToRiscvPass(ModulePass):
    name = "convert-scf-to-riscv-scf"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    ScfYieldLowering(),
                    ScfForLowering(),
                ]
            )
        ).rewrite_module(op)



"""
Dialect to generate C/C++ from MLIR.

The EmitC dialect allows to convert operations from other MLIR dialects to EmitC ops.
Those can be translated to C/C++ via the Cpp emitter.

See external [documentation](https://mlir.llvm.org/docs/Dialects/EmitC/).
"""

from collections.abc import Iterable, Sequence
from typing import cast

from xdsl.dialects.builtin import (
    ArrayAttr,
    BFloat16Type,
    ContainerType,
    Float16Type,
    Float32Type,
    Float64Type,
    FloatAttr,
    IndexType,
    IntAttr,
    IntegerAttr,
    IntegerType,
    ShapedType,
    StringAttr,
    TensorType,
    TupleType,
)
from xdsl.ir import (
    Attribute,
    AttributeCovT,
    Dialect,
    ParametrizedAttribute,
    SSAValue,
    TypeAttribute,
)
from xdsl.irdl import (
    IRDLOperation,
    ParsePropInAttrDict,
    irdl_attr_definition,
    irdl_op_definition,
    opt_prop_def,
    prop_def,
    var_operand_def,
    var_result_def,
)
from xdsl.parser import AttrParser
from xdsl.printer import Printer
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import isa


@irdl_attr_definition
class EmitC_ArrayType(
    ParametrizedAttribute, TypeAttribute, ShapedType, ContainerType[AttributeCovT]
):
    """EmitC array type"""

    name = "emitc.array"

    shape: ArrayAttr[IntAttr]
    element_type: AttributeCovT

    def __init__(
        self,
        shape: Iterable[int | IntAttr],
        element_type: AttributeCovT,
    ):
        shape = ArrayAttr(
            [IntAttr(dim) if isinstance(dim, int) else dim for dim in shape]
        )
        super().__init__(shape, element_type)

    def verify(self) -> None:
        if not self.shape.data:
            raise VerifyException("EmitC array shape must not be empty")

        for dim_attr in self.shape.data:
            if dim_attr.data < 0:
                raise VerifyException(
                    "EmitC array dimensions must have non-negative size"
                )

        element_type = self.get_element_type()

        if isinstance(element_type, EmitC_ArrayType):
            raise VerifyException(
                "EmitC array element type cannot be another EmitC_ArrayType."
            )

        # Check that the element type is a supported EmitC type.
        if not self._is_valid_element_type(element_type):
            raise VerifyException(
                f"EmitC array element type '{element_type}' is not a supported EmitC type."
            )

    def get_num_dims(self) -> int:
        return len(self.shape.data)

    def get_shape(self) -> tuple[int, ...]:
        return tuple(i.data for i in self.shape.data)

    def get_element_type(self) -> AttributeCovT:
        return self.element_type

    @classmethod
    def parse_parameters(cls, parser: AttrParser):
        with parser.in_angle_brackets():
            shape, type = parser.parse_ranked_shape()
            return ArrayAttr(IntAttr(dim) for dim in shape), type

    def print_parameters(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            printer.print_list(
                self.shape, lambda dim: printer.print_string(f"{dim.data}"), "x"
            )
            printer.print_string("x")
            printer.print_attribute(self.element_type)

    def _is_valid_element_type(self, element_type: Attribute) -> bool:
        """
        Check if the element type is valid for EmitC_ArrayType.
        See external [documentation](https://github.com/llvm/llvm-project/blob/main/mlir/include/mlir/Dialect/EmitC/IR/EmitCTypes.td#L77).
        """
        return is_integer_index_or_opaque_type(element_type) or is_supported_float_type(
            element_type
        )


@irdl_attr_definition
class EmitC_LValueType(ParametrizedAttribute, TypeAttribute):
    """
    EmitC lvalue type.
    Values of this type can be assigned to and their address can be taken.
    See [tablegen definition](https://github.com/llvm/llvm-project/blob/main/mlir/include/mlir/Dialect/EmitC/IR/EmitCTypes.td#L87)
    """

    name = "emitc.lvalue"
    value_type: TypeAttribute

    def verify(self) -> None:
        """
        Verify the LValueType.
        See [MLIR implementation](https://github.com/llvm/llvm-project/blob/main/mlir/lib/Dialect/EmitC/IR/EmitC.cpp#L1095)
        """
        # Check that the wrapped type is valid. This especially forbids nested lvalue types.
        if not is_supported_emitc_type(self.value_type):
            raise VerifyException(
                f"!emitc.lvalue must wrap supported emitc type, but got {self.value_type}"
            )
        if isinstance(self.value_type, EmitC_ArrayType):
            raise VerifyException("!emitc.lvalue cannot wrap !emitc.array type")


@irdl_attr_definition
class EmitC_OpaqueType(ParametrizedAttribute, TypeAttribute):
    """EmitC opaque type"""

    name = "emitc.opaque"
    value: StringAttr

    def verify(self) -> None:
        if not self.value.data:
            raise VerifyException("expected non empty string in !emitc.opaque type")
        if self.value.data[-1] == "*":
            raise VerifyException(
                "pointer not allowed as outer type with !emitc.opaque, use !emitc.ptr instead"
            )


@irdl_attr_definition
class EmitC_PointerType(ParametrizedAttribute, TypeAttribute):
    """EmitC pointer type"""

    name = "emitc.ptr"
    pointee_type: TypeAttribute

    def verify(self) -> None:
        if isinstance(self.pointee_type, EmitC_LValueType):
            raise VerifyException("pointers to lvalues are not allowed")


@irdl_attr_definition
class EmitC_PtrDiffT(ParametrizedAttribute, TypeAttribute):
    """
    EmitC signed pointer diff type.
    Signed data type as wide as platform-specific pointer types. In particular, it is as wide as emitc.size_t.
    It corresponds to ptrdiff_t found in <stddef.h>.
    """

    name = "emitc.ptrdiff_t"


@irdl_attr_definition
class EmitC_SignedSizeT(ParametrizedAttribute, TypeAttribute):
    """
    EmitC signed size type.
    Data type representing all values of emitc.size_t, plus -1. It corresponds to ssize_t found in <sys/types.h>.
    Use of this type causes the code to be non-C99 compliant.
    """

    name = "emitc.ssize_t"


@irdl_attr_definition
class EmitC_SizeT(ParametrizedAttribute, TypeAttribute):
    """
    EmitC unsigned size type.
    Unsigned data type as wide as platform-specific pointer types. It corresponds to size_t found in <stddef.h>.
    """

    name = "emitc.size_t"


_SUPPORTED_BITWIDTHS = (1, 8, 16, 32, 64)


def _is_supported_integer_type(type_attr: Attribute) -> bool:
    """
    Check if an IntegerType is supported by EmitC.
    See external [documentation](https://github.com/llvm/llvm-project/blob/main/mlir/lib/Dialect/EmitC/IR/EmitC.cpp#L96).
    """
    return (
        isinstance(type_attr, IntegerType)
        and type_attr.width.data in _SUPPORTED_BITWIDTHS
    )


def is_supported_float_type(type_attr: Attribute) -> bool:
    """
    Check if a type is a supported floating-point type in EmitC.
    See external [documentation](https://github.com/llvm/llvm-project/blob/main/mlir/lib/Dialect/EmitC/IR/EmitC.cpp#L117)
    """
    match type_attr:
        case Float16Type() | BFloat16Type() | Float32Type() | Float64Type():
            return True
        case _:
            return False


def is_pointer_wide_type(type_attr: Attribute) -> bool:
    """Check if a type is a pointer-wide type."""
    match type_attr:
        case EmitC_PtrDiffT() | EmitC_SignedSizeT() | EmitC_SizeT():
            return True
        case _:
            return False


def is_integer_index_or_opaque_type(
    type_attr: Attribute,
) -> bool:
    """
    Check if a type is an integer, index, or opaque type.

    The emitC opaque type is not implemented yet so this function currently checks
    only for integer and index types.
    See external [documentation](https://github.com/llvm/llvm-project/blob/main/mlir/lib/Dialect/EmitC/IR/EmitC.cpp#L112).
    """
    return (
        _is_supported_integer_type(type_attr)
        or isinstance(type_attr, IndexType)
        or is_pointer_wide_type(type_attr)
    )


def is_supported_emitc_type(type_attr: Attribute) -> bool:
    """
    Check if a type is supported by EmitC.
    See [MLIR implementation](https://github.com/llvm/llvm-project/blob/main/mlir/lib/Dialect/EmitC/IR/EmitC.cpp#L62).
    """
    match type_attr:
        case IntegerType():
            return _is_supported_integer_type(type_attr)
        case IndexType():
            return True
        case EmitC_OpaqueType():
            return True
        case EmitC_ArrayType():
            elem_type = cast(Attribute, type_attr.get_element_type())
            return not isinstance(
                elem_type, EmitC_ArrayType
            ) and is_supported_emitc_type(elem_type)
        case EmitC_PointerType():
            return is_supported_emitc_type(type_attr.pointee_type)
        case Float16Type() | BFloat16Type() | Float32Type() | Float64Type():
            return True
        case TensorType():
            elem_type = cast(Attribute, type_attr.get_element_type())
            if isinstance(elem_type, EmitC_ArrayType):
                return False
            return is_supported_emitc_type(elem_type)
        case TupleType():
            return all(
                not isinstance(t, EmitC_ArrayType) and is_supported_emitc_type(t)
                for t in type_attr.types
            )
        case EmitC_PtrDiffT():
            return True
        case _:
            return False


@irdl_op_definition
class EmitC_CallOpaqueOp(IRDLOperation):
    """
    The `emitc.call_opaque` operation represents a C++ function call. The callee can be an arbitrary non-empty string.
    The call allows specifying order of operands and attributes in the call as follows:

        - integer value of index type refers to an operand;
        - attribute which will get lowered to constant value in call;
    """

    name = "emitc.call_opaque"

    callee = prop_def(StringAttr)
    args = opt_prop_def(ArrayAttr)
    template_args = opt_prop_def(ArrayAttr)
    # The SSA‐value operands of the call
    call_args = var_operand_def()
    res = var_result_def()

    irdl_options = [ParsePropInAttrDict()]
    assembly_format = (
        "$callee `(` $call_args `)` attr-dict `:` functional-type(operands, results)"
    )

    def __init__(
        self,
        callee: StringAttr | str,
        call_args: Sequence[SSAValue],
        result_types: Sequence[Attribute],
        args: ArrayAttr[Attribute] | None = None,
        template_args: ArrayAttr[Attribute] | None = None,
        attributes: dict[str, Attribute] | None = None,
    ):
        if isinstance(callee, str):
            callee = StringAttr(callee)
        super().__init__(
            properties={
                "callee": callee,
                "args": args,
                "template_args": template_args,
            },
            operands=[call_args],
            result_types=[result_types],
            attributes=attributes,
        )

    def verify_(self) -> None:
        if not self.callee.data:
            raise VerifyException("callee must not be empty")

        if self.args is not None:
            for arg in self.args.data:
                if isa(arg, IntegerAttr[IndexType]):
                    index = arg.value.data
                    if not (0 <= index < len(self.call_args)):
                        raise VerifyException("index argument is out of range")
                elif isinstance(arg, ArrayAttr):
                    # see https://github.com/llvm/llvm-project/blob/2eb733b5a6ab17a3ae812bb55c1c7c64569cadcd/mlir/lib/Dialect/EmitC/IR/EmitC.cpp#L342
                    # This part is referenced as a FIXME there.
                    raise VerifyException("array argument has no type")

        if self.template_args is not None:
            for t_arg in self.template_args.data:
                if not isa(
                    t_arg,
                    TypeAttribute | IntegerAttr | FloatAttr,
                    # FIXME: uncomment and replace the line above when EmitC_OpaqueAttr is implemented
                    # TypeAttribute | IntegerAttr | FloatAttr | EmitC_OpaqueAttr,
                ):
                    raise VerifyException("template argument has invalid type")

        for res_type in self.res.types:
            if isinstance(res_type, EmitC_ArrayType):
                raise VerifyException("cannot return array type")


EmitC = Dialect(
    "emitc",
    [
        EmitC_CallOpaqueOp,
    ],
    [
        EmitC_ArrayType,
        EmitC_LValueType,
        EmitC_OpaqueType,
        EmitC_PointerType,
        EmitC_PtrDiffT,
        EmitC_SignedSizeT,
        EmitC_SizeT,
    ],
)



from dataclasses import dataclass

from typing_extensions import override

from xdsl.backend.register_stack import RegisterStack
from xdsl.dialects.riscv import FloatRegisterType, IntRegisterType


@dataclass
class RiscvRegisterStack(RegisterStack):
    """
    Available RISCV-specific registers.
    """

    DEFAULT_ALLOCATABLE_REGISTERS = (
        *reversed(IntRegisterType.allocatable_registers()),
        *reversed(FloatRegisterType.allocatable_registers()),
    )

    @classmethod
    @override
    def default_allocatable_registers(cls):
        return RiscvRegisterStack.DEFAULT_ALLOCATABLE_REGISTERS



"""
The Data Layout and Target Information (DLTI) dialect is intended to
hold attributes and other components pertaining to descriptions of
in-memory data layout and compilation targets.

https://mlir.llvm.org/docs/Dialects/DLTIDialect/
"""

from __future__ import annotations

from abc import ABC
from collections.abc import Mapping
from typing import TypeAlias

from xdsl.dialects.builtin import (
    ArrayAttr,
    IntAttr,
    StringAttr,
)
from xdsl.ir import Attribute, Dialect, ParametrizedAttribute, TypeAttribute
from xdsl.irdl import irdl_attr_definition, param_def
from xdsl.parser import AttrParser
from xdsl.printer import Printer
from xdsl.utils.exceptions import VerifyException

DictValueType: TypeAlias = Mapping[
    StringAttr | TypeAttribute | str, "Attribute | str | int | DictValueType"
]


class DLTITypeConverters:
    """
    Type converter for DLTIEntryMap and DataLayoutEntryAttr parameters
    """

    @staticmethod
    def convert_entry_map_type(
        contents: ArrayAttr[DataLayoutEntryAttr] | DictValueType,
    ) -> ArrayAttr[DataLayoutEntryAttr]:
        if isinstance(contents, Mapping):
            return ArrayAttr([DataLayoutEntryAttr(k, v) for k, v in contents.items()])
        else:
            return contents

    @staticmethod
    def convert_entry_attr_key_type(
        key: StringAttr | TypeAttribute | str,
    ) -> Attribute:
        if isinstance(key, str):
            return StringAttr(key)
        else:
            return key

    @staticmethod
    def convert_entry_attr_value_type(
        value: Attribute | str | int | DictValueType,
    ) -> Attribute:
        if isinstance(value, str):
            return StringAttr(value)
        elif isinstance(value, int):
            return IntAttr(value)
        elif isinstance(value, Mapping):
            return MapAttr(value)
        else:
            return value


@irdl_attr_definition
class DataLayoutEntryAttr(ParametrizedAttribute):
    """
    An attribute to represent an entry of a data layout specification.
    https://mlir.llvm.org/docs/Dialects/DLTIDialect/#datalayoutentryattr
    """

    name = "dlti.dl_entry"

    key: Attribute = param_def(converter=DLTITypeConverters.convert_entry_attr_key_type)
    value: Attribute = param_def(
        converter=DLTITypeConverters.convert_entry_attr_value_type
    )

    def verify(self) -> None:
        if not isinstance(self.key, StringAttr | TypeAttribute):
            raise VerifyException("key must be a string or a type attribute")
        if isinstance(self.key, StringAttr) and not self.key.data:
            raise VerifyException("empty string as DLTI key is not allowed")


class DLTIEntryMap(ParametrizedAttribute, ABC):
    """
    Many DLTI dialect operations contain arrays of DataLayoutEntryInterface,
    with these representing different things such as data layout or information
    about the target hardware. This is the base class that these operations extend.
    """

    # In MLIR, this is a DataLayoutEntryInterface.
    entries: ArrayAttr[DataLayoutEntryAttr] = param_def(
        converter=DLTITypeConverters.convert_entry_map_type
    )

    def __getitem__(self, key: StringAttr | TypeAttribute | str) -> Attribute:
        if isinstance(key, str):
            key = StringAttr(key)
        for entry in self.entries:
            if entry.key == key:
                return entry.value

        raise KeyError

    @classmethod
    def parse_parameters(
        cls, parser: AttrParser
    ) -> tuple[ArrayAttr[DataLayoutEntryAttr]]:
        def parse_entry() -> DataLayoutEntryAttr:
            entry = parser.parse_attribute()
            parser.parse_punctuation("=")
            value = parser.parse_attribute()
            return DataLayoutEntryAttr.new((entry, value))

        entries = parser.parse_comma_separated_list(parser.Delimiter.ANGLE, parse_entry)

        return (ArrayAttr(entries),)

    def print_parameters(self, printer: Printer):
        def print_entry(entry: DataLayoutEntryAttr):
            printer.print_attribute(entry.key)
            printer.print_string(" = ")
            printer.print_attribute(entry.value)

        with printer.in_angle_brackets():
            printer.print_list(self.entries, print_entry)

    def verify(self) -> None:
        if len({entry.key for entry in self.entries}) != len(self.entries):
            raise VerifyException("duplicate DLTI entry key")


@irdl_attr_definition(init=False)
class DataLayoutSpecAttr(DLTIEntryMap):
    """
    An attribute to represent a data layout specification.

    A data layout specification is a list of entries that specify (partial) data
    layout information. It is expected to be attached to operations that serve as
    scopes for data layout requests.

    https://mlir.llvm.org/docs/Dialects/DLTIDialect/#datalayoutspecattr
    """

    name = "dlti.dl_spec"


@irdl_attr_definition(init=False)
class TargetDeviceSpecAttr(DLTIEntryMap):
    """
    An attribute to represent target device specification.

    Each device specification describes a single device and its hardware properties.
    Each device specification can contain any number of optional hardware properties
    (e.g., max_vector_op_width below).

    https://mlir.llvm.org/docs/Dialects/DLTIDialect/#targetdevicespecattr
    """

    name = "dlti.target_device_spec"


@irdl_attr_definition(init=False)
class TargetSystemSpecAttr(DLTIEntryMap):
    """
    An attribute to represent target system specification.

    A system specification describes the overall system containing multiple devices,
    with each device having a unique ID (string) and its corresponding
    TargetDeviceSpec object.

    https://mlir.llvm.org/docs/Dialects/DLTIDialect/#targetsystemspecattr
    """

    name = "dlti.target_system_spec"


@irdl_attr_definition(init=False)
class MapAttr(DLTIEntryMap):
    """
    A mapping of DLTI-information by way of key-value pairs

    A Data Layout and Target Information map is a list of entries effectively
    encoding a dictionary, mapping DLTI-related keys to DLTI-related values.

    https://mlir.llvm.org/docs/Dialects/DLTIDialect/#mapattr
    """

    name = "dlti.map"


DLTI = Dialect(
    "dlti",
    [],
    [
        DataLayoutEntryAttr,
        DataLayoutSpecAttr,
        MapAttr,
        TargetDeviceSpecAttr,
        TargetSystemSpecAttr,
    ],
)
